function Project(env::Environment, prms)

  @unpack mesh = env

  if isperiodic(mesh)
    error("GRHD project does not support periodic meshes!")
  end

  formulation = prms["GRHD"]["formulation"]
  if mesh isa Mesh1d
    if !(formulation in ["valencia1d", "spherical1d", "rescaled_spherical1d", "doublecartoon"])
      TODO("formulation = $formulation")
    end
  elseif mesh isa Mesh2d
    if !(formulation in ["cartoon"])
      TODO("Only cartoon formulation is supported")
    end
  end

  problem = if startswith(prms["GRHD"]["id"], "bondi_")
    if formulation == "doublecartoon"
      error("GRHD with doublecartoon formulation currently only supports TOV initial data!")
    end
    :bondi
  elseif prms["GRHD"]["id"] == "tov"
    :tov
  end

  atm_equalize_on_interface = prms["GRHD"]["atm_equalize_on_interface"]
  av_drag = prms["GRHD"]["av_drag"]
  av_sensor_abslog_D = Bool(prms["GRHD"]["av_sensor_abslog_D"])
  slope_limiter_method = Symbol(prms["GRHD"]["slope_limiter_method"])
  slope_limiter_tvb_M = prms["GRHD"]["slope_limiter_tvb_M"]
  c2p_dynamic_atm = prms["GRHD"]["c2p_dynamic_atm"]
  bernstein = BernsteinReconstruction(env.mesh)
  cold_K, cold_Γ = prms["GRHD"]["c2p_cold_eos_parameters"]
  cold_eos = Polytrope(cold_K, cold_Γ)
  c2p_set_atmosphere_on_failure = prms["GRHD"]["c2p_set_atmosphere_on_failure"]
  c2p_enforce_atm = Bool(prms["GRHD"]["c2p_enforce_atm"])
  c2p_enforce_causal_atm = Bool(prms["GRHD"]["c2p_enforce_causal_atm"])
  atm_evolve = Bool(prms["GRHD"]["atm_evolve"])
  hrsc = Symbol(prms["GRHD"]["hrsc"])
  freeze_vars = Symbol.(prms["GRHD"]["freeze_vars"])
  fv_numerical_flux = prms["GRHD"]["fv_numerical_flux"]
  id = prms["GRHD"]["id"]
  bc = prms["GRHD"]["bc"]
  workspace = dg1d.Workspace()
  if mesh.element isa DGElement
    if mesh.element.kind === :modal_bspline2
      V_bspline = dg1d.Bspline.vandermonde_matrix(mesh.element.z, mesh.element.bs2)
      invV_bspline = pinv(V_bspline)
    elseif mesh.element.kind === :modal_bspline1
      V_bspline = dg1d.Bspline.vandermonde_matrix(mesh.element.z, mesh.element.bs1)
      invV_bspline = pinv(V_bspline)
    else
      V_bspline, invV_bspline = nothing, nothing
    end
  else
    V_bspline, invV_bspline = nothing, nothing
  end
  fixedprms = (; av_regularization=:covariant, id_smooth=true,
                 bernstein, slope_limiter_method, slope_limiter_tvb_M,
                 c2p_dynamic_atm, atm_evolve, atm_equalize_on_interface,
                 c2p_set_atmosphere_on_failure, c2p_enforce_causal_atm, c2p_enforce_atm,
                 av_drag, av_sensor_abslog_D, problem, freeze_vars,
                 V_bspline, invV_bspline, workspace,
                 id, bc, hrsc, fv_numerical_flux)

  # construct Project
  eos = EquationOfState.make_EquationOfState(prms["EquationOfState"])
  equation = make_Equation(eos, cold_eos, prms["GRHD"], env.mesh)
  hrsc, tci = if mesh isa Mesh1d
    hrsc = HRSC.make_HRSC(env.mesh, prms["HRSC"])
    tci = TCI.make_TCI(env.mesh, prms["TCI"])
    hrsc, tci
  else
    nothing, nothing
  end
  P = Project(Symbol(formulation), equation, hrsc, tci, fixedprms, prms)

  register_all!(env.mesh, P)
  display(env.mesh.cache)

  for var in prms["GRHD"]["freeze_vars"]
    if formulation != "doublecartoon" && formulation != "spherical1d"
      TODO("GRHD.freeze_vars option not yet implemented for this formulation")
    end
    if !isregistered(env.mesh.cache, Symbol(var), :dynamic_variables)
      error("GRHD.freeze_vars: requested to freeze $var, but this is not an evolved variable!")
    end
  end

  initialdata!(env, P)
  projectcb = make_callback(env, P)
  append!(env.callbacks, CallbackSet(projectcb.callbacks))

  return P, nothing
end


#######################################################################
#                         Register variables                          #
#######################################################################


function register_all!(mesh, P::Project)
  register_evolution!(mesh, P)
  register_metricvars!(mesh, P)
  register_hrsc!(mesh, P, P.hrsc)
  register_tci!(mesh, P, P.tci)
  register_analysis!(mesh, P)

  vars0d = P.prmsdb["GRHD"]["variables0d_analyze_error"]
  vars1d = P.prmsdb["GRHD"]["variables1d_analyze_error"]
  refsol = P.prmsdb["GRHD"]["analyze_error_reference_solution"]
  if (!isempty(vars0d) || !isempty(vars1d)) && isempty(refsol)
    error("""requested error analysis variables with GRHD.variables0d_analyze_error or GRHD.variables1d_analyze_error,
          but no choice of reference solution was made with GRHD.analyze_error_reference_solution""")
  end
  if !isempty(refsol)
    if !isempty(vars0d)
      register_variables!(mesh, static_variablenames=(:ana_err,))
    end
    for var in vars0d
      if !isregistered(mesh.cache, Symbol(var), :static_variables) &&
            !isregistered(mesh.cache, Symbol(var), :dynamic_variables)
        error("parameter GRHD.variables0d_analyze_error: $var is not a static or dynamic variable!")
      end
      var_ref = string(var, "_ref")
      var_err_norm = string(var, "_err_norm")
      register_variables!(mesh,
                          static_variablenames=(Symbol(var_ref),),
                          global_variablenames=(Symbol(var_err_norm),)
                         )
    end
    for var in vars1d
      if !isregistered(mesh.cache, Symbol(var), :static_variables) &&
            !isregistered(mesh.cache, Symbol(var), :dynamic_variables)
        error("parameter GRHD.variables1d_analyze_error: $var is not a static or dynamic variable!")
      end
      var_ref = string(var, "_ref")
      if !isregistered(mesh.cache, Symbol(var_ref)) # might have been registered with vars0d above
        register_variables!(mesh, static_variablenames=(Symbol(var_ref),))
      end
      var_err = string(var, "_err")
      register_variables!(mesh, static_variablenames=(Symbol(var_err),))
    end
  end
end


#######################################################################
#                         evolution variables                         #
#######################################################################


function register_evolution!(mesh, P::Project{:valenciad1d})
  register_variables!(mesh,
    dynamic_variablenames = (:rD, :rSr, :rτ,),
    rhs_variablenames     = (:rhs_rD, :rhs_rSr, :rhs_rτ), # rhs of conservatives
    static_variablenames  = (:flr_rD, :flr_rSr, :flr_rτ,  # conservatives' fluxes
                             :src_rD, :src_rSr, :src_rτ,  # sources
                             :D, :Sr, :τ,
                             :ρ, :vr, :p, :ϵ,          # primitives
                             :max_v, :r,               # maximum charactersitic speed, radius
                             :init_D, :init_Sr, :init_τ, :init_max_v, :init_vr, :init_p,
                             :init_ρ, :init_ϵ,
                             :E, :Em1, :flr_E, :flr_Em1, :EP,
                             :ρhW2),
    bdry_variablenames    = (:bdry_D, :bdry_Sr, :bdry_τ,
                             :bdry_ρ, :bdry_ϵ, :bdry_vr, :bdry_p,
                             :bdry_max_v,
                             :nflr_rD, :nflr_rSr, :nflr_rτ),
    cell_variablenames    = (:cellmax_v,),
    global_variablenames  = (:t, :tm1) # time steps
  )
  @unpack r, x = get_static_variables(mesh)
  r .= x
end

function register_evolution!(mesh, P::Project{:spherical1d})
  register_variables!(mesh,
    dynamic_variablenames = (:D, :Sr, :τ,),         # scale conservatives
    rhs_variablenames     = (:rhs_D, :rhs_Sr, :rhs_τ), # rhs of conservatives
    static_variablenames  = (:flr_D, :flr_Sr, :flr_τ,  # conservatives' fluxes
                             :src_D, :src_Sr, :src_τ,  # sources
                             :ρ, :vr, :p, :ϵ,                 # primitives
                             :raw_ρ, :raw_vr, :raw_p, :raw_ϵ, # unlimited primitives
                             :max_v, :r,               # maximum charactersitic speed, radius
                             :init_D, :init_Sr, :init_τ, :init_max_v, :init_vr, :init_p,
                             :init_ρ, :init_ϵ,
                             :E, :Em1, :flr_E, :flr_Em1, :EP,
                             :ρhW2),
    bdry_variablenames    = (:bdry_D, :bdry_Sr, :bdry_τ,
                             :bdry_ρ, :bdry_ϵ, :bdry_vr, :bdry_p,
                             :bdry_max_v,
                             :nflr_D, :nflr_Sr, :nflr_τ),
    cell_variablenames    = (:cellmax_v,),
    global_variablenames  = (:t, :tm1) # time steps
  )
  @unpack r, x = get_static_variables(mesh)
  r .= x
  register_variables!(mesh,
    static_variablenames = (:D_modal, :Sr_modal, :τ_modal,
                            :flr_D_modal, :flr_Sr_modal, :flr_τ_modal,
                            :src_D_modal, :src_Sr_modal, :src_τ_modal,
                            :rhs_D_modal, :rhs_Sr_modal, :rhs_τ_modal,
                            :c2p_failed)
  )
end
function register_evolution!(mesh::Mesh1d{FVElement}, P::Project{:spherical1d})
  hrsc = P.prmsdb["GRHD"]["hrsc"]
  if hrsc === "fv_central"
    register_variables!(mesh,
      dynamic_variablenames = (:D, :Sr, :τ,),
      rhs_variablenames     = (:rhs_D, :rhs_Sr, :rhs_τ), # rhs of conservatives
      static_variablenames  = (:flr_D, :flr_Sr, :flr_τ,  # conservatives' fluxes
                               :src_D, :src_Sr, :src_τ,  # sources
                               :ρ, :vr, :p, :ϵ,          # primitives
                               :max_v, :r,               # maximum charactersitic speed, radius
                               :init_D, :init_Sr, :init_τ, :init_max_v, :init_vr, :init_p,
                               :init_ρ, :init_ϵ,
                               :E, :Em1, :flr_E, :flr_Em1, :EP,
                               :ρhW2),
      bdry_variablenames    = (:bdry_D, :bdry_Sr, :bdry_τ,
                               :bdry_ρ, :bdry_ϵ, :bdry_vr, :bdry_p,
                               :bdry_max_v,
                               :nflr_D, :nflr_Sr, :nflr_τ),
      cell_variablenames    = (:cellmax_v,),
      global_variablenames  = (:t, :tm1) # time steps
    )
  elseif hrsc === "slope_limiter"
    register_variables!(mesh,
      dynamic_variablenames = (:D, :Sr, :τ,),
      rhs_variablenames     = (:rhs_D, :rhs_Sr, :rhs_τ), # rhs of conservatives
      static_variablenames  = (:bD, :bSr, :bτ,           # buffers
                               :D1, :Sr1, :τ1,
                               :D2, :Sr2, :τ2,
                               :flr_D, :flr_Sr, :flr_τ,
                               :src_D, :src_Sr, :src_τ,  # sources
                               # we are slope limiting primitive variables after recovery
                               # following suggestion in Thierfelder et al.
                               :ρL, :vrL, :ϵL, :pL,      # slope limited values
                               :ρR, :vrR, :ϵR, :pR,      # slope limited values
                               :dρ_p, :dvr_p, :dϵ_p,     # diffs left
                               :dρ_m, :dvr_m, :dϵ_m,     # diffs right
                               :dρ_L, :dvr_L, :dϵ_L,     # limited diffs
                               :nflx_D_L, :nflx_Sr_L, :nflx_τ_L,
                               :nflx_D_R, :nflx_Sr_R, :nflx_τ_R,
                               :ρ, :vr, :p, :ϵ,          # primitives
                               :max_v, :r,               # maximum charactersitic speed, radius
                               :E, :Em1, :flr_E, :flr_Em1, :EP,
                               :init_D, :init_Sr, :init_τ, :init_max_v, :init_vr, :init_p,
                               :init_ρ, :init_ϵ,
                               :ρhW2),
      global_variablenames  = (:t, :tm1) # time steps
    )
  else
    TODO(hrsc)
  end
  @unpack r, x = get_static_variables(mesh)
  r .= x
end


function register_evolution!(mesh, P::Project{:rescaled_spherical1d})
  register_variables!(mesh,
    dynamic_variablenames = (:rD, :rSr, :rτ,),            # we evolve rescaled variables
    rhs_variablenames     = (:rhs_rD, :rhs_rSr, :rhs_rτ), # rhs of conservatives
    static_variablenames  = (:flr_rD, :flr_rSr, :flr_rτ,  # conservatives' fluxes
                             :src_rD, :src_rSr, :src_rτ,  # sources
                             :D, :Sr, :τ,                 # conserved variables
                             :ρ, :vr, :p, :ϵ,             # primitives
                             :max_v, :r,                  # maximum charactersitic speed, radius
                             :init_D, :init_Sr, :init_τ, :init_max_v, :init_vr, :init_p,
                             :init_ρ, :init_ϵ,
                             :E, :Em1, :flr_E, :flr_Em1, :EP,
                             :ρhW2),
    bdry_variablenames    = (:bdry_D, :bdry_Sr, :bdry_τ,
                             :bdry_ρ, :bdry_ϵ, :bdry_vr, :bdry_p,
                             :bdry_max_v,
                             :nflr_rD, :nflr_rSr, :nflr_rτ),
    cell_variablenames    = (:cellmax_v,),
    global_variablenames  = (:t, :tm1) # time steps
  )
  @unpack r, x = get_static_variables(mesh)
  r .= x
end
function register_evolution!(mesh::Mesh1d{FVElement}, P::Project{:rescaled_spherical1d})
  TODO()
end


function register_evolution!(mesh, P::Project{:doublecartoon})
  register_variables!(mesh,
    dynamic_variablenames = (:rD, :rSx, :rτ), # we evolve rescaled variables
    rhs_variablenames     = (:rhs_rD, :rhs_rSx, :rhs_rτ), # rhs of conservatives
    static_variablenames  = (:flx_rDx, :∂yflx_rDy, :∂zflx_rDz,
                             :flx_rSxx, :∂yflx_rSxy, :∂zflx_rSxz,
                             :flx_rτx, :∂yflx_rτy, :∂zflx_rτz,
                             :buf_cartoon,
                             :src_rD, :src_rSx, :src_rSy, :src_rSz, :src_rτ,  # sources
                             :D, :Sx, :τ,       # conserved variables
                             :ρ, :vx, :p, :ϵ,   # primitives
                             :max_v, :r,                  # maximum charactersitic speed, radius
                             :E, :Em1, :flr_E, :flr_Em1, :EP, # TODO Still needed?
                             :ρhW2),
    bdry_variablenames    = (:bdry_D, :bdry_Sx, :bdry_τ,
                             :bdry_ρ, :bdry_ϵ, :bdry_vx, :bdry_p,
                             :bdry_max_v,
                             :nflx_rD, :nflx_rSx, :nflx_rτ),
    cell_variablenames    = (:cellmax_v,),
    global_variablenames  = (:t, :tm1) # time steps
  )
  @unpack r, x = get_static_variables(mesh)
  @. r = x
end
function register_evolution!(mesh::Mesh1d{FVElement}, P::Project{:doublecartoon})
  hrsc = P.prmsdb["GRHD"]["hrsc"]
  if hrsc === "fv_central"
    register_variables!(mesh,
      dynamic_variablenames = (:rD, :rSx, :rτ,),
      rhs_variablenames     = (:rhs_rD, :rhs_rSx, :rhs_rτ), # rhs of conservatives
      static_variablenames  = (:flx_rDx, :flx_rSxx, :flx_rτx,  # conservatives' fluxes
                               :∂yflx_rDy, :∂zflx_rDz,
                               :∂yflx_rSxy, :∂zflx_rSxz,
                               :∂yflx_rτy, :∂zflx_rτz,
                               :src_rD, :src_rSx, :src_rτ,  # sources
                               :ρ, :vx, :p, :ϵ,          # primitives
                               :D, :Sx, :τ,
                               :max_v, :r,                # maximum charactersitic speed, radius
                               :init_D, :init_Sx, :init_τ, :init_max_v, :init_vx, :init_p,
                               :init_ρ, :init_ϵ,
                               :buf_cartoon, :ρhW2),
      bdry_variablenames    = (:bdry_rD, :bdry_rSx, :bdry_rτ,
                               :bdry_ρ, :bdry_ϵ, :bdry_vx, :bdry_p,
                               :bdry_max_v,
                               :nflx_rD, :nflx_rSx, :nflx_rτ),
      cell_variablenames    = (:cellmax_v,),
      global_variablenames  = (:t, :tm1) # time steps
    )
  elseif hrsc == "slope_limiter"
    register_variables!(mesh,
      dynamic_variablenames = (:rD, :rSx, :rτ), # we evolve rescaled variables
      rhs_variablenames     = (:rhs_rD, :rhs_rSx, :rhs_rτ), # rhs of conservatives
      static_variablenames  = (:flx_rDx, :∂yflx_rDy, :∂zflx_rDz,
                               :flx_rSxx, :∂yflx_rSxy, :∂zflx_rSxz,
                               :flx_rτx, :∂yflx_rτy, :∂zflx_rτz,
                               :buf_cartoon,
                               :brD, :brSx, :brτ,           # buffers
                               :rD1, :rSx1, :rτ1,
                               :rD2, :rSx2, :rτ2,
                               :ρL, :vxL, :ϵL, :pL,      # slope limited values
                               :ρR, :vxR, :ϵR, :pR,      # slope limited values
                               :dρ_p, :dvx_p, :dϵ_p,     # diffs left
                               :dρ_m, :dvx_m, :dϵ_m,     # diffs right
                               :dρ_L, :dvx_L, :dϵ_L,     # limited diffs
                               :dρ_R, :dvx_R, :dϵ_R,     # limited diffs
                               :nflx_rD_L, :nflx_rSx_L, :nflx_rτ_L,
                               :nflx_rD_R, :nflx_rSx_R, :nflx_rτ_R,
                               :src_rD, :src_rSx, :src_rSy, :src_rSz, :src_rτ,  # sources
                               :D, :Sx, :τ,       # conserved variables
                               :ρ, :vx, :p, :ϵ,   # primitives
                               :max_v, :r,                # maximum charactersitic speed, radius
                               :E, :Em1, :flr_E, :flr_Em1, :EP, # TODO Still needed?
                               :ρhW2),
      cell_variablenames    = (:cellmax_v,),
      global_variablenames  = (:t, :tm1) # time steps
    )
  else
    TODO(hrsc)
  end
  @unpack r, x = get_static_variables(mesh)
  @. r = x
end


function register_evolution!(mesh::Mesh2d, P::Project{:cartoon})
  register_variables!(mesh,
    dynamic_variablenames = (:rD, :rSx, :rSz, :rτ), # we evolve rescaled variables
    rhs_variablenames     = (:rhs_rD, :rhs_rSx, :rhs_rSz, :rhs_rτ), # rhs of conservatives
    static_variablenames  = (:flx_rDx,  :∂yflx_rDy,  :flx_rDz,
                             :flx_rSxx, :∂yflx_rSxy, :flx_rSxz,
                             :flx_rSzx, :∂yflx_rSzy, :flx_rSzz,
                             :flx_rτx,  :∂yflx_rτy,  :flx_rτz,
                             :buf_cartoon,
                             :src_rD, :src_rSx, :src_rSy, :src_rSz, :src_rτ,  # sources
                             :D, :Sx, :Sz, :τ,       # conserved variables
                             :ρ, :vx, :vz, :p, :ϵ,   # primitives
                             :max_v, :max_vx, :max_vz, :r, # maximum charactersitic speed, radius
                             :ρhW2),
    bdry_variablenames    = (:bdry_D, :bdry_Sx, :bdry_Sz, :bdry_τ,
                             :bdry_ρ, :bdry_ϵ, :bdry_vx, :bdry_vz, :bdry_p,
                             :bdry_max_v,
                             :nflx_rD, :nflx_rSx, :nflx_rSz, :nflx_rτ),
    cell_variablenames    = (:cellmax_v,),
    global_variablenames  = (:t, :tm1) # time steps
  )
  @unpack r, x, y = get_static_variables(mesh)
  @. r = sqrt(x^2+y^2)
end


#######################################################################
#                          metric variables                           #
#######################################################################


function register_metricvars!(mesh, P::Project{:valencia1d})
  register_variables!(mesh,
    static_variablenames = (:α, :βru,
                            :γrr, :γθθ, :γϕϕ, :sqrt_detγ,
                            :Krr, :Kθθ, :Kϕϕ,
                            :Γrrr, :Γrθθ, :Γrϕϕ, :Γθrθ, :Γθϕϕ, :Γϕrϕ, :Γϕθϕ,
                            :∂αr, :∂βrrdu,
                            :∂γrrr, :∂γrθθ, :∂γrϕϕ, :∂γθϕϕ,)
  )
end

function register_metricvars!(mesh, P::Project{:spherical1d})
  register_variables!(mesh,
    static_variablenames = (:α, :βru,
                            :γrr, :γθθ, :γϕϕ,
                            :Krr, :Kθθ, :Kϕϕ,
                            :Γrrr, :Γrθθ, :Γrϕϕ, :Γθrθ, :Γθϕϕ, :Γϕrϕ, :Γϕθϕ,
                            :∂αr, :∂βrrdu,
                            :∂γrrr, :∂γrθθ, :∂γrϕϕ, :∂γθϕϕ,)
  )
end

function register_metricvars!(mesh, P::Project{:rescaled_spherical1d})
  register_variables!(mesh,
    static_variablenames = (:A, :∂Ar, :B, :∂Br, :γrr)
  )
end

function register_metricvars!(mesh, P::Project{:doublecartoon})
  register_variables!(mesh,
    static_variablenames = (:α, :βux, :βuy, :βuz,
                            :∂αx, :∂αy, :∂αz,
                            :∂βuxx, :∂βuyx, :∂βuzx,
                            :∂βuxy, :∂βuyy, :∂βuzy,
                            :∂βuxz, :∂βuyz, :∂βuzz,
                            :ψ4, :∂ψ4x,
                            :γxx, :γxy, :γxz,
                            :γyy, :γyz,
                            :γzz,
                            :sqrtdetγ,
                            :Kxx, :Kxy, :Kxz,
                            :Kyy, :Kyz,
                            :Kzz,
                            :∂γxxx, :∂γxxy, :∂γxxz,
                            :∂γxyy, :∂γxyz,
                            :∂γxzz,
                            :∂γyxx, :∂γyxy, :∂γyxz,
                            :∂γyyy, :∂γyyz,
                            :∂γyzz,
                            :∂γzxx, :∂γzxy, :∂γzxz,
                            :∂γzyy, :∂γzyz,
                            :∂γzzz,
                           )
  )
end

function register_metricvars!(mesh::Mesh2d, P::Project{:cartoon})
  register_variables!(mesh,
    static_variablenames = (:α, :βux, :βuy, :βuz,
                            :∂αx, :∂αy, :∂αz,
                            :∂βuxx, :∂βuyx, :∂βuzx,
                            :∂βuxy, :∂βuyy, :∂βuzy,
                            :∂βuxz, :∂βuyz, :∂βuzz,
                            :ψ4, :∂ψ4x, :∂ψ4z,
                            :γxx, :γxy, :γxz,
                            :γyy, :γyz,
                            :γzz,
                            :sqrtdetγ,
                            :Kxx, :Kxy, :Kxz,
                            :Kyy, :Kyz,
                            :Kzz,
                            :∂γxxx, :∂γxxy, :∂γxxz,
                            :∂γxyy, :∂γxyz,
                            :∂γxzz,
                            :∂γyxx, :∂γyxy, :∂γyxz,
                            :∂γyyy, :∂γyyz,
                            :∂γyzz,
                            :∂γzxx, :∂γzxy, :∂γzxz,
                            :∂γzyy, :∂γzyz,
                            :∂γzzz,
                           )
  )
end


#######################################################################
#                         analysis variables                          #
#######################################################################


# TODO Where to put these? Can be needed by either HRSC or TCI.
# :E, :Em1,                   # entropy
# :flx_E, :flx_Em1,           # entropy flux

function register_analysis!(mesh::Mesh1d, P)
  register_variables!(mesh,
      static_variablenames = (:c2p_reset_ϵ, :c2p_reset_atm, :c2p_limit_vr,
                              :c2p_freeze_atm, :c2p_init_admissible, :v),
      bdry_variablenames = (:bdry_c2p_reset_atm,),
      cell_variablenames = (:flag,),
  )
  if "Mtot" in P.prmsdb["GRHD"]["variables0d_analyze"]
    register_variables!(mesh, global_variablenames = (:Mtot,))
  end
end
function register_analysis!(mesh::Mesh2d, P)
  register_variables!(mesh,
      static_variablenames = (:c2p_reset_ϵ, :c2p_reset_atm, :c2p_limit_vr,
                              :c2p_freeze_atm, :c2p_init_admissible, :v,
                              :buf_c2p_reset_atm,),
      bdry_variablenames = (:bdry_c2p_reset_atm,)
  )
  if "Mtot" in P.prmsdb["GRHD"]["variables0d_analyze"]
    TODO()
    register_variables!(mesh, global_variablenames = (:Mtot,))
  end
end


#######################################################################
#                                HRSC                                 #
#######################################################################

register_hrsc!(mesh, P, hrsc::Nothing)                     = nothing
register_hrsc!(mesh, P, hrsc::HRSC.AbstractReconstruction) = nothing
register_hrsc!(mesh, P, hrsc::AbstractHRSC)                = TODO(hrsc)

register_hrsc!(mesh, P, hrsc::HRSC.BernsteinReconstruction) = nothing

function register_hrsc!(mesh, P::Project{:doublecartoon}, hrsc::HRSC.BernsteinReconstruction)
  register_variables!(mesh,
    static_variablenames  = (:log_abs_rD,)
  )
end

function register_hrsc!(mesh, P::Project{:valencia1d}, hrsc::HRSC.AbstractArtificialViscosity)
  register_variables!(mesh,
    static_variablenames  = (:ldg_rD, :ldg_rSr, :ldg_rτ,
                             :av_u),
    bdry_variablenames    = (:bdry_ldg_rD, :bdry_ldg_rSr, :bdry_ldg_rτ,
                             :bdry_smoothed_mu,
                             :bdry_rhs_rD, :bdry_rhs_rSr, :bdry_rhs_rτ),
    cell_variablenames    = (:mu, :mu_m1, :tmp_mu), # 'one viscosity to rule them all'
    )
  if hrsc isa HRSC.SmoothedArtificialViscosity
    register_variables!(mesh,
      static_variablenames  = (:smoothed_mu,)
    )
  end
end

function register_hrsc!(mesh, P::Project{:spherical1d}, hrsc::HRSC.AbstractArtificialViscosity)
  register_variables!(mesh,
    static_variablenames  = (:ldg_D, :ldg_Sr, :ldg_τ,
                             :av_u),
    bdry_variablenames    = (:bdry_ldg_D, :bdry_ldg_Sr, :bdry_ldg_τ,
                             :bdry_smoothed_mu,
                             :bdry_rhs_D, :bdry_rhs_Sr, :bdry_rhs_τ),
    cell_variablenames    = (:mu, :mu_m1, :tmp_mu), # 'one viscosity to rule them all'
    )
  if hrsc isa HRSC.SmoothedArtificialViscosity
    register_variables!(mesh,
      static_variablenames  = (:smoothed_mu,)
    )
  end
end

function register_hrsc!(mesh, P::Project{:doublecartoon}, hrsc::HRSC.AbstractArtificialViscosity)
  register_variables!(mesh,
    static_variablenames  = (:ldg_rD, :ldg_rSx, :ldg_rτ,
                             :av_u),
    bdry_variablenames    = (:bdry_ldg_rD, :bdry_ldg_rSx, :bdry_ldg_rτ,
                             :bdry_smoothed_mu,
                             :bdry_rhs_rD, :bdry_rhs_rSx, :bdry_rhs_rτ),
    cell_variablenames    = (:mu, :mu_m1, :tmp_mu), # 'one viscosity to rule them all'
    )
  if hrsc isa HRSC.SmoothedArtificialViscosity
    register_variables!(mesh,
      static_variablenames  = (:smoothed_mu,)
    )
  end
end


#######################################################################
#                                 TCI                                 #
#######################################################################


register_tci!(mesh, P, tci::Nothing) = nothing
register_tci!(mesh, P, tci::TCI.AbstractTCI) = TODO(tci)


function register_tci!(mesh, P::Project{:spherical1d}, tci::TCI.AbstractTCI)
  TODO(tci)
  register_variables!(mesh, cell_variablenames = (:flag,:sD_flag,:sSr_flag,:stau_flag))
end

function register_tci!(mesh, P::Project{:spherical1d}, tci::TCI.EntropyProduction)
  TODO(tci)
  register_variables!(mesh,
                      cell_variablenames = (:flag,),
                      static_variablenames = (:EP,), # entropy production
                      bdry_variablenames   = (:EPjump_lhs, :EPjump_rhs)
                                              # entropy production due to jumps at interfaces
  )
end

function register_tci!(mesh, P::Project{:doublecartoon}, tci::TCI.AbstractTCI)
  register_variables!(mesh, cell_variablenames = (:flag,))
end
